import abc
import base64
import datetime
import functools
import hashlib
import hmac
import io
import json
import os
import requests

from collections import OrderedDict
from typing import Callable, NewType, Optional
from urllib.parse import quote, urlparse

from connectors_core.azure import ConvertRestToken, AzureAuthenticationType

ConvertedRestToken = NewType("ConvertedRestToken", dict)


class ADLSFileBase(io.RawIOBase):
    def __init__(self, requests, path, headers, debug, mode):
        if mode != "rb":
            raise NotImplementedError(f"File mode must be 'rb' not {mode}")
        self.requests = requests
        self.path = path
        self.mode = mode
        self.headers = headers
        self.pos = 0
        self.debug = debug
        self.fsize = self._get_file_size()

    def tell(self) -> int:
        """ Current file offset """
        return self.pos

    def seekable(self):
        return True

    def readable(self):
        return True

    def seek(self, offset, whence=os.SEEK_SET):
        self.debug(f"ADLSFile.seek / offset: {offset}, whence: {whence}")
        if whence == os.SEEK_SET:
            self.pos = offset
        elif whence == os.SEEK_CUR:
            self.pos += offset
        elif whence == os.SEEK_END:
            self.pos = self.fsize + offset
        else:
            raise NotImplementedError(f"Unsuporrted whence: {whence}")
        return self.pos

    def read(self, n=-1):
        if self.pos >= self.fsize:
            return b""
        # x-ms-range: bytes=start=end header to specify len
        headers = self.headers
        if n >= 0:
            headers["x-ms-range"] = f"bytes={self.pos}-{self.pos+n-1}"
        else:
            headers["x-ms-range"] = f"bytes={self.pos}-"
        self.debug(f'ADLSFile.read / Reading {headers["x-ms-range"]}')
        r = self.requests.get(self.path, headers=headers)
        r.raise_for_status()
        self.pos += int(r.headers["Content-Length"])
        self.debug(f"data: {r.content[:10]}")
        return r.content

    def readinto(self, b):
        output = self.read(len(b))
        length = len(output)
        b[:length] = output
        return length

    def _get_file_size(self):
        r = self.requests.head(self.path, headers=self.headers)
        r.raise_for_status()
        metadata = dict(r.headers)
        try:
            return int(metadata["Content-Length"])
        except KeyError:
            raise FileNotFoundError()


class ADLSFile(io.BufferedReader):
    BUFFER_SIZE = 50_000_000

    def __init__(self, requests, path, headers, debug, mode="rb"):
        if mode not in {"rb", "wb", "ab"}:
            raise NotImplementedError(
                f"File mode must be {{'rb', 'wb', 'ab'}} not {mode}"
            )
        base_stream = ADLSFileBase(requests, path, headers, debug, mode)
        self.fsize = base_stream.fsize
        super().__init__(base_stream, buffer_size=self.BUFFER_SIZE)


def wrap_requests_errors(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except requests.exceptions.HTTPError as e:
            if e.response.headers.get("x-ms-error-code") == "PathNotFound":
                file_path_idx = e.args[0].find("url:")
                if file_path_idx != -1:
                    raise FileNotFoundError(e.args[0][file_path_idx + len("url:") :])
                raise FileNotFoundError(e.args[0])
            raise ConnectionError(e.args[0])

    return wrapper


class AzureRequests:
    def __init__(self, instance):
        self.instance = instance

    def _request(self, request, *args, **kwargs):
        try:
            r = request(*args, **kwargs)
            r.raise_for_status()
            return r
        except requests.exceptions.HTTPError as e:
            if (
                e.response.headers.get("x-ms-error-code") == "InvalidAuthenticationInfo"
            ):  # refresh token
                self.instance._token = (
                    self.instance.refresh()
                )  # refreshes the access token
                if "headers" not in kwargs:
                    kwargs["headers"] = dict()
                kwargs["headers"].update(self.instance.authorization_header)
                return request(*args, **kwargs)
            raise  # reraise the exception if it's not authorization error

    def __getattr__(self, name):
        """
        Check if caller tries to invoke function from requests library and if so,
        it returns our wrapper which handles token refreshing
        If attribute is not a function in requests, it raises AttributeError as expected
        """

        if hasattr(requests, name) and callable(getattr(requests, name)):
            return functools.partial(self._request, getattr(requests, name))
        raise AttributeError(
            f"type object '{type(self).__name__}' has no attribute '{name}'"
        )


class AzureRequestsSharedKey(AzureRequests):
    API_VERSION = "2018-03-28"

    def canonize_headers(self, headers):
        return "\n".join(
            map(
                lambda x: f"{x[0].lower()}:{x[1]}",
                sorted(filter(lambda x: x[0].startswith("x-ms-"), headers.items())),
            )
        )

    def urlencode(self, url):
        o = urlparse(url)
        encoded_url = o.scheme + "://" + o.netloc
        path = url[len(encoded_url) :]
        # ~ should not be escaped, but we need to specify it manually
        # because of bug in urllib for <3.7
        return encoded_url + quote(path, "/~")

    def canonize_resources(self, url, params):
        path = urlparse(url).path.lstrip("/")

        result_str = "/" + self.instance.storage_account_name + "/" + path

        if params:
            return (
                result_str
                + "\n"
                + "\n".join(
                    map(lambda x: f"{x[0].lower()}:{x[1]}", sorted(params.items()))
                )
            )
        return result_str

    @wrap_requests_errors
    def _request(self, request, *args, **kwargs):
        if "headers" in kwargs:
            headers = kwargs["headers"]
        else:
            headers = dict()
        if "params" in kwargs:
            params = kwargs["params"]
        else:
            params = dict()
        if "data" in kwargs:
            data = kwargs["data"]
        else:
            data = b""
        url = self.urlencode(args[0])
        new_headers = self._prepare_headers(
            request.__name__, url, headers, params, data
        )
        kwargs["headers"] = new_headers
        r = request(url, *args[1:], **kwargs)
        r.raise_for_status()
        return r

    def _prepare_headers(
        self, request: str, url, headers: dict = None, params: dict = None, data=b""
    ):

        if "x-ms-date" not in headers:
            headers["x-ms-date"] = datetime.datetime.utcnow().strftime(
                "%a, %d %b %Y %H:%M:%S GMT"
            )
        if "x-ms-version" not in headers:
            headers["x-ms-version"] = self.API_VERSION

        content_length = headers.get("Content-Length", "")
        if content_length == "0":
            content_length = ""

        # Ordering is guaranteed from 3.7, so we need to ensure it's working on 3.6 as intended
        params = OrderedDict(
            (
                ("verb", request.upper()),
                ("Content-Encoding", headers.get("Content-Encoding", "")),
                ("Content-Language", headers.get("Content-Language", "")),
                ("Content-Length", content_length),
                ("Content-MD5", headers.get("Content-MD5", "")),
                ("Content-Type", headers.get("Content-Type", "application/json")),
                ("Date", headers.get("Date", "")),
                ("If-Modified-Since", headers.get("If-Modified-Since", "")),
                ("If-Match", headers.get("If-Match", "")),
                ("If-None-Match", headers.get("If-None-Match", "")),
                ("If-Unmodified-Since", headers.get("If-Unmodified-Since", "")),
                ("Range", headers.get("Range", "")),
                ("CanonicalizedHeaders", self.canonize_headers(headers)),
                ("CanonicalizedResources", self.canonize_resources(url, params)),
            )
        )

        res_str = "\n".join(params.values())

        signature_string = base64.b64encode(
            hmac.new(
                base64.b64decode(self.instance.shared_key),
                msg=res_str.encode("utf-8"),
                digestmod=hashlib.sha256,
            ).digest()
        ).decode()

        new_headers = {
            "x-ms-date": headers["x-ms-date"],
            "x-ms-version": headers["x-ms-version"],
            "Content-Type": headers.get("Content-Type", "application/json"),
            "Content-Length": headers.get("Content-Length", "0"),
            "Authorization": (
                "SharedKey "
                + self.instance.storage_account_name
                + ":"
                + signature_string
            ),
        }

        for key, value in headers.items():
            if key.startswith("x-ms-"):
                new_headers[key] = value

        return new_headers


class ADLSGen2:
    CHUNK_SIZE = 2 ** 24  # default buffer size for a streaming buffer, 16 MiB

    def __init__(
        self,
        authority_uri: str,
        storage_account_name: str,
        dns_suffix: str,
        file_system: str,
        token: ConvertedRestToken,
        engine: object,
        debug: Callable[..., None] = lambda *args: None,
        buffer_size: Optional[int] = None,
    ):
        self.authority_uri = authority_uri.rstrip("/")
        self.storage_account_name = storage_account_name
        self.dns_suffix = dns_suffix
        self.file_system = file_system
        self.debug = debug
        self._token = token
        self.alteryx_engine = engine
        self.requests = AzureRequests(self)
        self.buffer = io.BytesIO()
        # buffer size is set in MB
        self.buffer_size = buffer_size * 2 ** 20 if buffer_size else self.CHUNK_SIZE
        self.debug(f"Buffer size = {self.buffer_size} bytes")

    @wrap_requests_errors
    def info(self, path: str) -> Optional[dict]:
        r = self.requests.head(f'{self.url}/{path.lstrip("/")}', headers=self.headers)
        r.raise_for_status()
        self.debug(f"info headers: {r.headers}")
        metadata = dict(r.headers)
        self.debug(f"{metadata}")
        # checks if file exits
        r.raise_for_status()
        metadata["length"] = int(metadata["Content-Length"])
        return metadata

    def seekable(self) -> bool:
        return False

    @wrap_requests_errors
    def open(self, path: str, mode: str = "rb") -> "ADLSGen2":
        self.base_url = f'{self.url}/{path.lstrip("/")}'
        self.debug(f"OPEN url: {self.base_url}")
        if mode == "rb":
            return ADLSFile(self.requests, self.base_url, self.headers, self.debug)
        elif mode == "wb":
            headers = self.headers
            headers["Content-Type"] = "application/octet-stream"
            r = self.requests.put(
                self.base_url,
                data=b"",
                headers=self.headers,
                params={"resource": "file"},
            )
            r.raise_for_status()
            self.debug(f"{r.text}")
            self.current_data_len = 0
        elif mode == "ab":
            meta = self.info(path)
            try:
                self.current_data_len = meta["length"]
            except (KeyError, TypeError):
                # file does not exists, create new
                return self.open(path, "wb")
        else:
            raise ValueError('Mode has to be "rb", "wb", or "ab"')
        return self

    def close(self) -> None:
        self.flush()
        self.base_url = ""
        self.current_data_len = 0

    def __enter__(self):
        return self.stream_content

    def __exit__(self, type, value, traceback):
        self.r.close()

    @wrap_requests_errors
    def _send_buffer(self) -> None:
        self.debug(f"Sending {self.buffer.tell()} bytes of data")
        headers = self.headers
        headers["Content-Type"] = "application/octet-stream"
        headers["Content-Length"] = str(self.buffer.tell())
        r = self.requests.patch(
            self.base_url,
            data=self.buffer.getbuffer(),
            headers=headers,
            params={"action": "append", "position": self.current_data_len},
        )
        r.raise_for_status()
        self.current_data_len += self.buffer.tell()
        self.debug(f"{r.headers}")
        self.buffer = io.BytesIO()

    @wrap_requests_errors
    def write(self, data: bytes) -> int:
        if not data:
            return 0
        self.buffer.write(data)

        if self.buffer.tell() >= self.buffer_size:
            self._send_buffer()
        return len(data)

    @wrap_requests_errors
    def flush(self) -> None:
        # rest of buffer
        if self.buffer.tell() > 0:
            self._send_buffer()

        headers = self.headers
        headers["Content-Type"] = "application/octet-stream"
        headers["Content-Length"] = "0"
        r = self.requests.patch(
            self.base_url,
            headers=headers,
            params={"action": "flush", "position": self.current_data_len},
        )
        r.raise_for_status()
        self.debug(f"{r.headers}")

    @property
    def authorization_header(self) -> dict:
        return {"Authorization": f'{self.token["tokenType"]} {self.token["access"]}'}

    @property
    def headers(self) -> dict:
        return self.authorization_header

    @property
    def url(self) -> str:
        return (
            f"https://{self.storage_account_name}.{self.dns_suffix}/"
            f"{self.file_system}"
        )

    @property
    def token(self) -> ConvertedRestToken:
        return self._token

    @token.setter
    def token(self, value) -> None:
        self._token = value

    @property
    def tenant_id(self) -> str:
        return self.token["tenant"]

    @property
    def client_id(self) -> str:
        return self.token["client"]

    @property
    def refresh_token(self) -> str:
        return self.token["refresh"]

    @property
    def secret(self) -> Optional[str]:
        return self.token.get("secret")

    @property
    def resource(self) -> Optional[str]:
        return self.token.get("resource")


class ADLSService(ADLSGen2):
    def refresh(self) -> ConvertedRestToken:
        url = f"{self.authority_uri}/{self.tenant_id}/oauth2/token"
        files = {
            "grant_type": (None, "client_credentials"),
            "client_id": (None, self.client_id),
            "client_secret": (None, self.secret),
            "resource": (None, self.resource),
        }

        r = requests.post(url, files=files)
        r.raise_for_status()
        token = r.json()
        token["not_before"] = 0
        return ConvertRestToken(
            token=token,
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.secret,
        )


class ADLSServiceEngine(ADLSService):
    def refresh(self) -> ConvertedRestToken:
        try:
            params = {
                "tenant_id": self.tenant_id,
                "client_id": self.client_id,
                "client_secret": self.secret,
                "authority_uri": self.authority_uri + "/",
                "resource_uri": self.resource,
                "flow": "client_credentials",
            }
            response = self.alteryx_engine.azure_authenticate(json.dumps(params))
            response = json.loads(response)

            if "error_code" in response:
                raise ConnectionError(response["error_code"])

            if "token_type" not in response:
                response["token_type"] = self.token["tokenType"]
            response["not_before"] = response["expires_on"]

            return ConvertRestToken(
                token=response,
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.secret,
            )

        except AttributeError:  # function is not exported in current version
            return super().refresh()


class ADLSEndUser(ADLSGen2):
    def refresh(self) -> ConvertedRestToken:
        url = f"{self.authority_uri}/{self.tenant_id}/oauth2/token"
        files = {
            "grant_type": (None, "refresh_token"),
            "client_id": (None, self.client_id),
            "refresh_token": (None, self.refresh_token),
        }

        if self.resource:
            files["resource"] = (None, self.resource)

        if self.secret:
            files["client_secret"] = (None, self.secret)

        r = requests.post(url, files=files)
        r.raise_for_status()
        token = r.json()
        token["not_before"] = token["expires_on"]
        return ConvertRestToken(
            token=token,
            tenant_id=self.tenant_id,
            client_id=self.client_id,
            client_secret=self.secret,
        )


class ADLSEndUserEngine(ADLSEndUser):
    def refresh(self) -> ConvertedRestToken:
        try:
            params = {
                "tenant_id": self.tenant_id,
                "client_id": self.client_id,
                "authority_uri": self.authority_uri + "/",
                "refresh_token": self.refresh_token,
            }
            if self.secret:
                params["client_secret"] = self.secret
            if self.resource:
                params["resource_uri"] = self.resource

            response = self.alteryx_engine.azure_refresh(json.dumps(params))
            response = json.loads(response)

            if "error_code" in response:
                raise ConnectionError(response["error_code"])

            # TODO fix engine to return token_type as well...
            if "token_type" not in response:
                response["token_type"] = self.token["tokenType"]
            response["not_before"] = response["expires_on"]

            return ConvertRestToken(
                token=response,
                tenant_id=self.tenant_id,
                client_id=self.client_id,
                client_secret=self.secret,
            )

        except AttributeError:  # function is not exported in current version
            return super().refresh()


class ADLSSharedKey(ADLSGen2):
    def __init__(self, *args, **kwargs):
        self.shared_key = kwargs.pop("shared_key")
        super().__init__(*args, **kwargs)
        self.requests = AzureRequestsSharedKey(self)

    @property
    def headers(self):
        return dict()


def get_adls(*args, auth_type, **kwargs):
    if auth_type == AzureAuthenticationType.SharedKey:
        return ADLSSharedKey(*args, **kwargs)

    kwargs.pop("shared_key")  # not needed for end user and service
    if auth_type == AzureAuthenticationType.EndUser:
        return ADLSEndUserEngine(*args, **kwargs)
    # no other option here, valididy of auth_type is checked in the caller
    return ADLSServiceEngine(*args, **kwargs)
