import AyxProxySession

import AlteryxPythonSDK as Sdk
from typing import IO, NewType, Optional, Union


from azure.datalake.store.exceptions import DatalakeRESTException

import connectors_core.alteryx as alteryx
from connectors_core.azure import (
    AzureAuthenticationType,
    AdlSpAuthentication,
    AdlEndUserAuthentication,
    DATA_LAKE_TOKEN_AUDIENCE,
    ConvertRestToken,
)
from connectors_core.file import (
    DelimitedReader,
    JsonPivotInputReader,
    AvroInputReader,
    FileFormat,
    ExcelReader,
)

from connectors_core.file.errors import DelimitedIndexError

from utils import (
    getAuthType,
    getDnsSuffix,
    getStorageVersion,
    getToken,
    StorageVersion,
    get_adl_client,
)

import json


ConvertedRestToken = NewType("ConvertedRestToken", dict)


class AyxPlugin(alteryx.AyxPluginBase):
    """
    Implements the plugin interface methods, to be utilized by the Alteryx engine to communicate with this plugin.
    Prefixed with "pi", the Alteryx engine will expect the below five interface methods to be defined.
    """

    def __init__(
        self,
        n_tool_id: int,
        alteryx_engine: Sdk.AlteryxEngine,
        output_anchor_mgr: Sdk.OutputAnchorManager,
    ):
        """
        Constructor is called whenever the Alteryx engine wants to instantiate an instance of this plugin.
        :param n_tool_id: The assigned unique identification for a tool instance.
        :param alteryx_engine: Provides an interface into the Alteryx engine.
        :param output_anchor_mgr: A helper that wraps the outgoing connections for a plugin.
        """
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)

        # Configuration Options
        self.adl_token = None
        self.adl_auth_type = AzureAuthenticationType.ServicePrincipal
        self.adl_store_name = None  # type: str
        self.adl_tenant_id = None  # type: str
        self.adl_client_id = None  # type: str
        self.adl_secret_key = None  # type: str
        self.adl_file_path = ""
        self.adl_file_format = FileFormat.Delimited
        self.adl_delimited_has_header = True
        self.adl_field_size = 255
        self.adl_field_delimiter = ","
        self.adl_field_quote_character = '"'
        self.adl_encoding = "utf-8"
        self.adl_json_force_string_values = True
        self.adl_start_import_at_line = 1

        self.adl_authentication = None
        self.adl_shared_key = None

    def load_xml_shared_key(self, str_xml):
        self.adl_shared_key = self.get_encrypted_xml_config_value(
            str_xml, "adl_StorageAccountKey"
        )
        if not self.adl_shared_key:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a shared key")
            )

    def load_xml_application_info(self, str_xml):
        self.adl_tenant_id = self.get_xml_config_value(str_xml, "adl_TenantId") or ""
        self.adl_client_id = self.get_xml_config_value(str_xml, "adl_ClientId") or ""
        self.adl_secret_key = (
            self.get_encrypted_xml_config_value(str_xml, "adl_SecretKey") or ""
        )
        if self.adl_auth_type == AzureAuthenticationType.EndUserBasic:
            # once we read proper client & tenant id, everything else works the same as EndUser Auth
            self.adl_auth_type = AzureAuthenticationType.EndUser
            self.adl_tenant_id = (
                self.get_xml_config_value(str_xml, "adl_AlteryxTenantId") or ""
            )
            self.adl_client_id = (
                self.get_xml_config_value(str_xml, "adl_AlteryxClientId") or ""
            )
            self.adl_secret_key = ""

        try:
            self.adl_token = getToken(
                self.get_encrypted_xml_config_value(str_xml, "adl_Token"),
                self.adl_tenant_id,
                self.adl_client_id,
                self.adl_secret_key,
                self.adl_storage_version,
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide an access token")
            )

        if not self.adl_tenant_id:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a Tenant ID")
            )
        if not self.adl_client_id:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a Client ID")
            )

    def pi_init(self, str_xml: str):
        """
        Handles extracting user-entered file name, and input data verification.
        Called when the Alteryx engine is ready to provide the tool configuration from the GUI.
        :param str_xml: The raw XML from the GUI.
        """

        # this will set is_initialized=True, which will be set to False if the config is invalid
        super().pi_init(str_xml)

        # Getting the user-entered file name string from the GUI, and the output anchor from the XML file.
        self.adl_store_name = self.get_xml_config_value(str_xml, "adl_StoreName")
        try:
            self.adl_storage_version = getStorageVersion(
                self.get_xml_config_value(str_xml, "adl_StorageVersion")
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify valid Storage Version")
            )

        try:
            self.adl_auth_type = getAuthType(
                self.get_xml_config_value(str_xml, "adl_AuthenticationType")
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify valid Authentication Type")
            )

        if self.adl_auth_type == AzureAuthenticationType.SharedKey:
            self.load_xml_shared_key(str_xml)
        else:
            self.load_xml_application_info(str_xml)

        self.adl_authority_uri = (
            self.get_xml_config_value(str_xml, "adl_Endpoint")
            or "https://login.microsoftonline.com"
        )

        try:
            self.adl_dns_suffix = getDnsSuffix(
                self.get_xml_config_value(str_xml, "adl_StorageEndpointSuffix"),
                self.adl_storage_version,
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg(
                    "Please specify valid DNS Suffix or Storage Version"
                )
            )

        self.file_system_name = (
            self.get_xml_config_value(str_xml, "adl_Gen2FileSystemName") or None
        )
        self.sheet_name = self.get_xml_config_value(str_xml, "adl_SheetName") or None

        # Validation
        if not self.adl_store_name:
            self.invalidate_config(
                self.alteryx_engine.xmsg(
                    "Please provide your Azure Data Lake Store Name."
                )
            )

        if self.adl_auth_type == AzureAuthenticationType.ServicePrincipal:
            # Validation (client secret is required for service principal)
            if not self.adl_secret_key:
                self.invalidate_config(
                    self.alteryx_engine.xmsg("Please provide a Secret Key")
                )

            if self.is_initialized:
                self.adl_authentication = AdlSpAuthentication(
                    store_name=self.adl_store_name,
                    tenant_id=self.adl_tenant_id,
                    client_id=self.adl_client_id,
                    secret_key=self.adl_secret_key,
                )
        elif self.adl_auth_type == AzureAuthenticationType.EndUser:
            if self.is_initialized:
                self.adl_authentication = AdlEndUserAuthentication(
                    tenant_id=self.adl_tenant_id, token=self.adl_token
                )
        elif self.adl_auth_type == AzureAuthenticationType.SharedKey:
            # self.adl_uathentication is not used for gen2
            pass
        else:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify valid Authentication Type")
            )

        # File Configuration
        self.adl_file_format = FileFormat[
            self.get_xml_config_value(str_xml, "adl_FileFormat")
            or FileFormat.Unknown.name
        ]
        self.adl_file_path = self.get_xml_config_value(str_xml, "adl_FilePath")
        self.adl_json_force_string_values = (
            self.get_xml_config_value(str_xml, "adl_JSONForceStringValues") == "True"
        )

        self.adl_delimited_has_header = (
            self.get_xml_config_value(str_xml, "adl_DelimitedHasHeader") == "True"
        )
        self.adl_start_import_at_line = int(
            self.get_xml_config_value(str_xml, "adl_StartImportAtLine") or 1
        )

        try:
            if self.adl_start_import_at_line < 1:
                self.invalidate_config(
                    self.alteryx_engine.xmsg(
                        "Invalid Starting Line: must be greater or equal to 1"
                    )
                )
        except ValueError:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Invalid Starting Line: must be a number")
            )

        try:
            self.adl_field_size = int(
                self.get_xml_config_value(str_xml, "adl_MaxFieldLength") or 255
            )
            if self.adl_field_size < 1:
                self.invalidate_config(
                    self.alteryx_engine.xmsg(
                        "Invalid Max Field Length: must be greater than zero"
                    )
                )
        except ValueError:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Invalid Max Field Length: must be a number")
            )

        try:
            self.adl_field_quote_character = {
                "QUOTE": "'",
                "DOUBLE_QUOTE": '"',
                "BACKTICK": "`",
            }[
                self.get_xml_config_value(str_xml, "adl_QuoteCharacter")
                or "DOUBLE_QUOTE"
            ]
        except KeyError:
            self.invalidate_config(self.alteryx_engine.xmsg("Invalid Quote Character"))

        self.adl_encoding = (
            self.get_xml_config_value(str_xml, "adl_CodePage") or "utf-8"
        )

        try:
            self.adl_field_delimiter = (
                self.get_xml_config_value(str_xml, "adl_Delimiter") or "COMMA"
            )
            self.adl_field_delimiter = self.adl_field_delimiter.strip(" ")
            self.adl_field_delimiter = {
                "COMMA": ",",
                "TAB": "\t",
                "SPACE": " ",
                "": " ",  # check for a case space has been inserted as a custom delimiter
                # There's a hack for reading csv line by line
                # we use bell character '\7' in a hope that it is not a
                # part of a text. It's not printable so it shouldn't be.
                "NEW_LINE": "\n",
                "NO_DELIMITER": "\7",
            }[self.adl_field_delimiter]
        except KeyError:
            if len(self.adl_field_delimiter) != 1:
                self.invalidate_config(
                    self.alteryx_engine.xmsg("Invalid Field Delimiter")
                )

        if not self.adl_file_path:
            # TODO: enforce starting file paths with / ?
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a valid File Path")
            )
        if self.adl_file_format == FileFormat.Unknown:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify a File Format")
            )

        try:
            self.adl_client = get_adl_client(self)
        except ConnectionError as e:
            self.invalidate_config(str(e))

        metadata_dict = {
            "Host": DATA_LAKE_TOKEN_AUDIENCE,
            "ServiceName": "Azure Data Lake Store",
            "PluginName": "Azure Data Lake Store",
        }
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)
        self.output_anchor = self.output_anchor_mgr.get_output_anchor(
            "Output"
        )  # type: Sdk.OutputAnchor

    def pi_add_incoming_connection(
        self, str_type: str, str_name: str
    ) -> Union[alteryx.IncomingConnectionBase, None]:
        """
        The IncomingInterface objects are instantiated here, one object per incoming connection, however since
        this tool does not accept an incoming connection, instantiation is not needed and "ii" methods won't be called.
        Called when the Alteryx engine is attempting to add an incoming data connection.
        :param str_type: The name of the input connection anchor, defined in the Config.xml file.
        :param str_name: The name of the wire, defined by the workflow author.
        :return: self.
        """

        self.display_error_msg(
            self.alteryx_engine.xmsg("This tool does not accept incoming connections")
        )

        return None

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        """
        Called when the Alteryx engine is attempting to add an outgoing data connection.
        :param str_name: The name of the output connection anchor, defined in the Config.xml file.
        :return: True signifies that the connection is accepted.
        """

        # self.display_info_msg("Added connection {}".format(str_name))

        return True

    def pi_push_all_records(self, n_record_limit: int = -1) -> bool:
        """
        Handles pushing records out to downstream tool(s).
        Called when a tool has no incoming data connection.
        :param n_record_limit: Set it to <0 for no limit, 0 for no records, and >0 to specify the number of records.
        :return: False if there are issues with the input data or if the workflow isn't being ran, otherwise True.
        """
        if not self.is_initialized:
            return False

        update_only = (
            self.alteryx_engine.get_init_var(self.n_tool_id, "UpdateOnly") == "True"
            or n_record_limit == 0
        )

        try:
            file_meta = self.adl_client.info(path=self.adl_file_path)
            # See here for file info properties:
            # http://hadoop.apache.org/docs/stable/hadoop-project-dist/hadoop-hdfs/WebHDFS.html#Status_of_a_FileDirectory
            file_length_bytes = int(file_meta["length"])

            with self.adl_client.open(
                path=self.adl_file_path, mode="rb"
            ) as input_stream:
                if self.adl_file_format == FileFormat.Delimited:
                    reader = DelimitedReader(
                        input_stream,
                        self.output_anchor,
                        self.alteryx_engine,
                        file_length_bytes,
                        self.get_field_source,
                        binary_input=True,
                        max_field_size=self.adl_field_size,
                        delimiter=self.adl_field_delimiter,
                        quotechar=self.adl_field_quote_character,
                        encoding=self.adl_encoding,
                        has_header_row=self.adl_delimited_has_header,
                        start_data_at_line=self.adl_start_import_at_line,
                    )

                elif self.adl_file_format == FileFormat.Avro:
                    reader = AvroInputReader(
                        input_stream,
                        self.output_anchor,
                        self.alteryx_engine,
                        file_length_bytes,
                        self.get_field_source,
                    )

                elif self.adl_file_format == FileFormat.JSON:
                    reader = JsonPivotInputReader(
                        input_stream,
                        self.output_anchor,
                        self.alteryx_engine,
                        file_length_bytes,
                        self.get_field_source,
                        force_string_values=self.adl_json_force_string_values,
                        max_field_size=self.adl_field_size,
                    )

                elif self.adl_file_format == FileFormat.XLSX:
                    reader = ExcelReader(
                        stream=input_stream,
                        output_anchor=self.output_anchor,
                        alteryx_engine=self.alteryx_engine,
                        stream_length_bytes=file_length_bytes,
                        field_source_resolver=self.get_field_source,
                        has_header_row=self.adl_delimited_has_header,
                        start_data_at_line=self.adl_start_import_at_line,
                        sheet_name=self.sheet_name,
                    )

                else:
                    self.display_error_msg(
                        self.alteryx_engine.xmsg("Unknown File Format!")
                    )
                    return False

                reader.init()
                if update_only:
                    self.output_anchor.close()  # Close outgoing connections.
                    return True

                record_count = reader.read(n_record_limit)

            self.display_info_msg(
                self.alteryx_engine.xmsg(
                    "@1 records were read from @2/@3",
                    str(record_count),
                    self.adl_store_name,
                    self.adl_file_path,
                )
            )
            self.output_anchor.update_progress(1.0)
            self.output_anchor.close()  # Close outgoing connections.
            return True
        except FileNotFoundError as e:
            self.display_error_msg(
                self.alteryx_engine.xmsg("Unable to open file: @1", str(e))
            )
        except ConnectionError as e:
            self.display_error_msg(
                self.alteryx_engine.xmsg("Error establishing connection: @1", str(e))
            )
        except DatalakeRESTException as e:
            self.display_error_msg(
                self.alteryx_engine.xmsg("Error establishing connection: @1", str(e))
            )
        except DelimitedIndexError:
            self.display_error_msg(self.alteryx_engine.xmsg("Malformed csv file"))
        except (UnicodeDecodeError, UnicodeError) as e:
            self.display_error_msg(
                self.alteryx_engine.xmsg(
                    "Decoding of the input file has failed: @1", str(e)
                )
            )
        except IndexError as e:
            self.display_error_msg(
                self.alteryx_engine.xmsg(
                    "Processing of the input file has failed: @1", str(e)
                )
            )
        return False

    def pi_close(self, b_has_errors: bool):
        super().pi_close(b_has_errors)
        if self.output_anchor:
            self.output_anchor.assert_close()  # Checks whether connections were properly closed.

    def report_input_stream_progress(self, stream: IO, stream_length):
        stream_progress = round(stream.tell() / stream_length, 2)
        if (stream_progress * 100) % 5 == 0:
            self.output_anchor.update_progress(stream_progress)

    def get_field_source(self, field: Sdk.Field):
        return "ADLS: https://{0}.azuredatalakestore.net{1}".format(
            self.adl_store_name, self.adl_file_path
        )
