from adal import AdalError
from ADLSGen2 import get_adls
import azure.datalake.store as adl_store
from connectors_core.azure import AzureAuthenticationType, ConvertRestToken
from enum import Enum
from typing import NewType, Optional

import datetime
import json
import time

ConvertedRestToken = NewType('ConvertedRestToken', dict)

class StorageVersion(Enum):
    Gen1 = 'Gen1'
    Gen2 = 'Gen2'

def getAuthType(auth_type_xml : Optional[str]) -> AzureAuthenticationType:
    auth_type_str = auth_type_xml or AzureAuthenticationType.Unknown.name
    try:
        auth_type = AzureAuthenticationType[auth_type_str]
        if auth_type == AzureAuthenticationType.Unknown:
            raise ValueError("Please specify an Authentication Type")
        return auth_type
    except KeyError:
        raise ValueError("Please specify valid Authentication Type")

def getDnsSuffix(dns_suffix_xml : Optional[str], storage_version : StorageVersion) -> str:
    if dns_suffix_xml:
        return dns_suffix_xml
    if storage_version == StorageVersion.Gen1:
        return "azuredatalakestore.net"
    if storage_version == StorageVersion.Gen2:
        return "dfs.core.windows.net"
    raise ValueError("Please specify valid DNS Suffix or Storage Version")
    
def getStorageVersion(storage_version_xml : Optional[str]) -> StorageVersion:
    storage_version_str = storage_version_xml or StorageVersion.Gen1.name
    try:
        return StorageVersion[storage_version_str]
    except KeyError:
        raise ValueError("Please specify valid Storage Version")

def getResource(storage_version : StorageVersion) -> str:
    if storage_version == StorageVersion.Gen1:
        return 'https://datalake.azure.net/'
    if storage_version == StorageVersion.Gen2:
        return 'https://storage.azure.com/'
    raise ValueError("Please specify valid Storage Version")

def getToken(token_xml : Optional[str], tenant_id : str, client_id : str, secret_key : str, storage_version : StorageVersion) -> ConvertedRestToken:
    try:
        token = json.loads(token_xml)
    except (json.decoder.JSONDecodeError, TypeError):
        raise ValueError("Please provide an Access Token")
    if "not_before" not in token:
        token["not_before"] = 0
    if "token_type" not in token:
        token["token_type"] = "Bearer"
    if "resource" not in token:
        token["resource"] = getResource(storage_version)
    adl_token = ConvertRestToken(
        token=token,
        tenant_id=tenant_id,
        client_id=client_id,
        client_secret=secret_key
    )
    adl_token["time"] = time.time()
    return adl_token


def get_adl_client(self: 'ADLSTool', no_seek: bool = False):
    """
    Initialize a new ADL File System Client
    :return: an Azure Data Lake File System Client
    """

    # Don't throw error, just return.
    # It's called again after anitialization
    if not self.is_initialized:
        return
    
    if self.adl_storage_version == StorageVersion.Gen1:
        return _get_adl_gen1_client(self, no_seek)
    else:  # gen 2
        return get_adls(
            auth_type=self.adl_auth_type,
            authority_uri=self.adl_authority_uri,
            storage_account_name=self.adl_store_name,
            dns_suffix=self.adl_dns_suffix,
            file_system=self.file_system_name,
            token=self.adl_token,
            shared_key=self.adl_shared_key,
            engine=self.alteryx_engine,
        )


def no_seek_wrapper(fun):
    def wrapper(*args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except ValueError as e:
            raise AttributeError from e
    return wrapper


def _get_adl_gen1_client(self, no_seek):
    try:
        credentials = self.adl_authentication.get_credentials()
        if "resource" not in credentials.token:
            credentials.token["resource"] = "https://datalake.azure.net/"
        file_system = adl_store.AzureDLFileSystem(
            store_name=self.adl_store_name, token=credentials
        )
        try:
            if (
                int(credentials.token["expires_on"])
                < time.time() - datetime.timedelta(minutes=10).total_seconds()
            ):
                
                self.display_info_msg(f'refresh token: {credentials.token}')
                file_system.azure.token.refresh_token(
                    authority=self.adl_authority_uri + "/"
                )
        # Refresh token in case expires_on is not defined or it is malformed
        except (KeyError, ValueError):
            file_system.azure.token.refresh_token(
                authority=self.adl_authority_uri + "/"
            )

        # patch seek so it throws AttributeError for openpyxl
        if no_seek and not hasattr(adl_store.core.AzureDLFile.seek, 'wrapped'):
            adl_store.core.AzureDLFile.seek.wrapped = True
            adl_store.core.AzureDLFile.seek = no_seek_wrapper(adl_store.core.AzureDLFile.seek)

        return file_system
    except AdalError as e:
        tagIndex = str(e).find("server response: ")
        if tagIndex >= 0:
            braceIndex = str(e).find(
                "{", tagIndex
            )  # start looking at the tagIndex position
            if braceIndex >= 0:
                try:
                    # try to convert to JSON, will throw if it is not a valid JSON string
                    jsonString = json.loads(
                        str(e)[braceIndex:]
                    )  # jump to the opening brace and take remainder of string
                    if jsonString:
                        e = (
                            jsonString["error_description"]
                            .replace("\r\n", " ")
                            .replace("\r", " ")
                            .replace("\n", " ")
                        )
                except:
                    # ignore exception, fall through and report the error
                    pass
        errorString = self.alteryx_engine.xmsg(
            "There was a problem authenticating with Azure Data Lake: "
        )
        raise ConnectionError(errorString + str(e))
