import AyxProxySession

import AlteryxPythonSDK as Sdk
from typing import Any, NewType, Optional, Union

import datetime
import time


import azure.datalake.store as adl_store
from adal import AdalError

import connectors_core.alteryx as alteryx
import connectors_core.azure as core_azure
from connectors_core.azure import (
    AzureAuthenticationType,
    DATA_LAKE_TOKEN_AUDIENCE,
    ConvertRestToken,
)
import connectors_core.azure.filewriters as filewriters
from connectors_core.file import FileFormat, FileExistsBehavior
from connectors_core.file.interfaces import FileWriter
from connectors_core.file.errors import FileOpenError, FileWriterAbortedError

from ADLSGen2 import get_adls
from utils import (
    getAuthType,
    getDnsSuffix,
    getStorageVersion,
    getToken,
    StorageVersion,
    get_adl_client,
)

import json


ConvertedRestToken = NewType("ConvertedRestToken", dict)


class AyxPlugin(alteryx.AyxPluginBase):
    def __init__(
        self,
        n_tool_id: int,
        alteryx_engine: Sdk.AlteryxEngine,
        output_anchor_mgr: Sdk.OutputAnchorManager,
    ):
        super().__init__(n_tool_id, alteryx_engine, output_anchor_mgr)

        # Configuration Options
        self.adl_auth_type = AzureAuthenticationType.ServicePrincipal
        self.adl_token = None
        self.adl_store_name = None  # type: str
        self.adl_tenant_id = None  # type: str
        self.adl_client_id = None  # type: str
        self.adl_secret_key = None  # type: str
        self.adl_encoding = None  # type: str
        self.adl_sheet_name = None
        self.adl_file_path = ""
        self.adl_file_format = FileFormat.Delimited
        self.adl_delimited_has_header = True
        self.adl_existing_file_behavior = FileExistsBehavior.Abort
        self.adl_buffer_size = None

        self.adl_shared_key = None
        self.adl_authentication = None
        self.incoming_connection_added = False

    def pi_add_incoming_connection(
        self, str_type: str, str_name: str
    ) -> Union[alteryx.IncomingConnectionBase, None]:
        if self.incoming_connection_added:
            self.display_error_msg(
                self.alteryx_engine.xmsg(
                    "Only one incoming connection can be accepted by this tool"
                )
            )
            return None
        connection_handler = AzureDataLakeIncomingConnectionHandler(self)
        self.incoming_connection_added = True
        return connection_handler

    def load_xml_shared_key(self, str_xml):
        self.adl_shared_key = self.get_encrypted_xml_config_value(
            str_xml, "adl_StorageAccountKey"
        )
        if not self.adl_shared_key:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a shared key")
            )

    def load_xml_application_info(self, str_xml):
        self.adl_tenant_id = self.get_xml_config_value(str_xml, "adl_TenantId") or ""
        self.adl_client_id = self.get_xml_config_value(str_xml, "adl_ClientId") or ""
        self.adl_secret_key = (
            self.get_encrypted_xml_config_value(str_xml, "adl_SecretKey") or ""
        )
        try:
            self.adl_buffer_size = int(
                self.get_xml_config_value(str_xml, "adl_BufferSize")
            )
            # allowed buffer size <1;256> MiB
            if not 1 <= self.adl_buffer_size <= 256:
                self.adl_buffer_size = None
        except (ValueError, TypeError):
            # TODO log error of invalid buffer size
            pass

        if self.adl_auth_type == AzureAuthenticationType.EndUserBasic:
            # once we read proper client & tenant id, everything else works the same as EndUser Auth
            self.adl_auth_type = AzureAuthenticationType.EndUser
            self.adl_tenant_id = (
                self.get_xml_config_value(str_xml, "adl_AlteryxTenantId") or ""
            )
            self.adl_client_id = (
                self.get_xml_config_value(str_xml, "adl_AlteryxClientId") or ""
            )
            self.adl_secret_key = ""

        try:
            self.adl_token = getToken(
                self.get_encrypted_xml_config_value(str_xml, "adl_Token"),
                self.adl_tenant_id,
                self.adl_client_id,
                self.adl_secret_key,
                self.adl_storage_version,
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide an access token")
            )

        if not self.adl_tenant_id:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a Tenant ID")
            )
        if not self.adl_client_id:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a Client ID")
            )

    def pi_init(self, str_xml: str):
        super().pi_init(str_xml)

        # Getting the user-entered file name string from the GUI, and the output anchor from the XML file.
        try:
            self.adl_auth_type = getAuthType(
                self.get_xml_config_value(str_xml, "adl_AuthenticationType")
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify valid Authentication Type")
            )

        try:
            self.adl_storage_version = getStorageVersion(
                self.get_xml_config_value(str_xml, "adl_StorageVersion")
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify valid Storage Version")
            )

        if self.adl_auth_type == AzureAuthenticationType.SharedKey:
            self.load_xml_shared_key(str_xml)
        else:
            self.load_xml_application_info(str_xml)

        self.adl_store_name = self.get_xml_config_value(str_xml, "adl_StoreName")
        self.adl_sheet_name = self.get_xml_config_value(str_xml, "adl_SheetName")
        self.adl_authority_uri = (
            self.get_xml_config_value(str_xml, "adl_Endpoint")
            or "https://login.microsoftonline.com"
        )

        try:
            self.adl_dns_suffix = getDnsSuffix(
                self.get_xml_config_value(str_xml, "adl_StorageEndpointSuffix"),
                self.adl_storage_version,
            )
        except ValueError as e:
            self.invalidate_config(
                self.alteryx_engine.xmsg(
                    "Please specify valid DNS Suffix or Storage Version"
                )
            )

        self.file_system_name = (
            self.get_xml_config_value(str_xml, "adl_Gen2FileSystemName") or None
        )

        # Validation
        if not self.adl_store_name:
            self.invalidate_config(
                self.alteryx_engine.xmsg(
                    "Please provide your Azure Data Lake Store Name."
                )
            )

        if self.adl_auth_type == AzureAuthenticationType.ServicePrincipal:
            # Validation (client secret is required for service principal)
            if not self.adl_secret_key:
                self.invalidate_config(
                    self.alteryx_engine.xmsg("Please provide a Secret Key")
                )

            if self.is_initialized:
                self.adl_authentication = core_azure.AdlSpAuthentication(
                    store_name=self.adl_store_name,
                    tenant_id=self.adl_tenant_id,
                    client_id=self.adl_client_id,
                    secret_key=self.adl_secret_key,
                )
        elif self.adl_auth_type == AzureAuthenticationType.EndUser:
            if self.is_initialized:
                self.adl_authentication = core_azure.AdlEndUserAuthentication(
                    tenant_id=self.adl_tenant_id, token=self.adl_token
                )
        elif self.adl_auth_type == AzureAuthenticationType.SharedKey:
            # self.adl_uathentication is not used for gen2
            pass
        else:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify valid Authentication Type")
            )
            return

        # File Configuration
        self.adl_file_format = FileFormat[
            self.get_xml_config_value(str_xml, "adl_FileFormat")
            or FileFormat.Unknown.name
        ]
        self.adl_file_path = self.get_xml_config_value(str_xml, "adl_FilePath")

        self.adl_existing_file_behavior = FileExistsBehavior[
            self.get_xml_config_value(str_xml, "adl_ExistingFileBehavior")
            or FileExistsBehavior.Abort.name
        ]

        self.adl_encoding = (
            self.get_xml_config_value(str_xml, "adl_CodePage") or "utf-8"
        )

        if not self.adl_file_path:
            # TODO: enforce starting file paths with / ?
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please provide a valid File Path")
            )
        if self.adl_file_format == FileFormat.Unknown:
            self.invalidate_config(
                self.alteryx_engine.xmsg("Please specify a File Format")
            )

        try:
            self.adl_fs = get_adl_client(self, no_seek=True)
        except ConnectionError as e:
            self.invalidate_config(str(e))

        metadata_dict = {
            "Host": DATA_LAKE_TOKEN_AUDIENCE,
            "ServiceName": "Azure Data Lake Store",
            "PluginName": "Azure Data Lake Store",
        }
        self.alteryx_engine.create_connect_metadata(self.n_tool_id, metadata_dict)

    def pi_add_outgoing_connection(self, str_name: str) -> bool:
        self.display_error_msg(
            self.alteryx_engine.xmsg("This tool does not provide an output connection")
        )
        return False

    def pi_push_all_records(self, n_record_limit: int) -> bool:
        # We do not have an output to push records to, we are an input-only tool
        return True


class AzureDataLakeIncomingConnectionHandler(alteryx.IncomingConnectionBase):
    def __init__(self, parent: AyxPlugin):
        super().__init__(parent)
        self.writer = None  # type: FileWriter

    def ii_init(self, record_info_in: Sdk.RecordInfo) -> bool:
        if not self.parent.is_initialized:
            return False
        try:
            file_format = self.parent.adl_file_format

            def conversion_error_handler(
                field: Sdk.Field, record: Sdk.RecordRef, e: BaseException
            ) -> Any:
                if field.get_as_string(record) is None:
                    self.parent.display_warning_msg(
                        self.parent.alteryx_engine.xmsg(
                            "Missing value for field @1 (@2): @3",
                            field.name,
                            str(field.type),
                            str(e),
                        )
                    )
                else:
                    self.parent.display_error_msg(
                        self.parent.alteryx_engine.xmsg(
                            "Data conversion error: encountered for field @1 (@2): @3",
                            field.name,
                            str(field.type),
                            str(e),
                        )
                    )
                return None

            if file_format == FileFormat.Delimited:
                self.writer = filewriters.ADLDelimitedFileWriter(
                    adl_fs=self.parent.adl_fs,
                    file_path=self.parent.adl_file_path,
                    write_header=self.parent.adl_delimited_has_header,
                    data_conversion_error_handler=conversion_error_handler,
                    file_exists_behavior=self.parent.adl_existing_file_behavior,
                    encoding=self.parent.adl_encoding,
                    alteryx_engine=self.parent.alteryx_engine,
                )
            elif file_format == FileFormat.Avro:
                self.writer = filewriters.ADLAvroFileWriter(
                    adl_fs=self.parent.adl_fs,
                    file_path=self.parent.adl_file_path,
                    data_conversion_error_handler=conversion_error_handler,
                    file_exists_behavior=self.parent.adl_existing_file_behavior,
                    alteryx_engine=self.parent.alteryx_engine,
                )
            elif file_format == FileFormat.JSON:
                self.writer = filewriters.ADLJSONFileWriter(
                    adl_fs=self.parent.adl_fs,
                    file_path=self.parent.adl_file_path,
                    data_conversion_error_handler=conversion_error_handler,
                    file_exists_behavior=self.parent.adl_existing_file_behavior,
                    alteryx_engine=self.parent.alteryx_engine,
                )
            elif file_format == FileFormat.XLSX:
                self.writer = filewriters.ADLExcelFileWriter(
                    adl_fs=self.parent.adl_fs,
                    file_path=self.parent.adl_file_path,
                    write_header=self.parent.adl_delimited_has_header,
                    data_conversion_error_handler=conversion_error_handler,
                    file_exists_behavior=self.parent.adl_existing_file_behavior,
                    encoding=self.parent.adl_encoding,
                    alteryx_engine=self.parent.alteryx_engine,
                    sheet_title=self.parent.adl_sheet_name,
                )
            else:
                # Unknown file format!
                self.is_initialized = False
                return False
            self.writer.init(record_info_in)
            self.is_initialized = True
            return True
        except ConnectionError as e:
            self.is_initialized = False
            self.parent.display_error_msg(
                self.parent.alteryx_engine.xmsg(
                    "Failed to initialize ADL Output: @1", str(e)
                )
            )
            return False
        except FileOpenError as e:
            self.is_initialized = False
            self.parent.display_error_msg(
                self.parent.alteryx_engine.xmsg(
                    "Failed to initialize ADL Output: @1", str(e)
                )
            )
            return False
        except AdalError as e:
            self.is_initialized = False
            tagIndex = str(e).find("server response: ")
            if tagIndex >= 0:
                braceIndex = str(e).find(
                    "{", tagIndex
                )  # start looking at the tagIndex position
                if braceIndex >= 0:
                    try:
                        # try to convert to JSON, will throw if it is not a valid JSON string
                        jsonString = json.loads(
                            str(e)[braceIndex:]
                        )  # jump to the opening brace and take remainder of string
                        if jsonString:
                            e = (
                                jsonString["error_description"]
                                .replace("\r\n", " ")
                                .replace("\r", " ")
                                .replace("\n", " ")
                            )
                    except:
                        # ignore exception, fall through and report the error
                        pass
            errorString = self.parent.alteryx_engine.xmsg(
                "There was a problem authenticating with Azure Data Lake: "
            )
            self.parent.display_error_msg(errorString + str(e))
            return False

    def ii_push_record(self, in_record: Sdk.RecordRef) -> bool:
        try:
            if not self.writer.is_open():
                self.writer.open()
            return self.writer.write(in_record)
        except FileOpenError as e:
            self.parent.display_error_msg(
                self.parent.alteryx_engine.xmsg(
                    "Failed to open file for writing: @1", str(e)
                )
            )
            return False
        except FileWriterAbortedError as e:
            self.parent.display_error_msg(
                self.parent.alteryx_engine.xmsg(
                    "The write operation was aborted: @1", str(e)
                )
            )
            return False

    def ii_update_progress(self, d_percent: float):
        try:
            self.parent.alteryx_engine.output_tool_progress(
                self.parent.n_tool_id, d_percent
            )
        except AttributeError:
            pass

    def ii_close(self):
        self.writer.close()
        self.writer = None
        self.is_initialized = False
